import pytest
from leidenmark import leiden_plus


def test_base():
    # Paragraphs/sections/textparts
    assert leiden_plus('no paragraph') == 'no paragraph'
    assert leiden_plus('paragraph', enable_paragraphs=True) == '<p>paragraph</p>'
    # Issue #1: assert leiden_plus('<= Leiden+ section =>', enable_paragraphs=True) == '<ab>Leiden+ section</ab>'
    assert leiden_plus('<= Leiden+ section =>') == '<ab>Leiden+ section</ab>'
    assert leiden_plus('<D=.1 Leiden+ textpart =D>') == '<div n="1" type="textpart">Leiden+ textpart</div>'


LEIDEN_PLUS_SAMPLE = """\
<D=.r<=
1. Lorem ipsum dolor
vac.1lin
2. sit amet, con[ca.3]c
3.-etur adipiscing
=>=D>
<D=.v<=
lost.2lin
6. ut labore et dol
7.-ore magna aliqua
=>=D>
"""

LEIDEN_PLUS_SAMPLE_OUTPUT = """\
<div n="r" type="textpart">
  <ab>
    <l n="1">Lorem ipsum dolor</l>
    <space quantity="1" unit="line"/>
    <l n="2">sit amet, con<gap precision="low" quantity="3" reason="lost" unit="character"/>c</l>
    <l break="no" n="3">etur adipiscing</l>
  </ab>
</div>
<div n="v" type="textpart">
  <ab>
    <gap quantity="2" unit="line"/>
    <l n="6">ut labore et dol</l>
    <l break="no" n="7">ore magna aliqua</l>
  </ab>
</div>\
"""


@pytest.fixture
def leiden_plus_sample():
    return LEIDEN_PLUS_SAMPLE


@pytest.fixture
def leiden_plus_sample_output():
    return LEIDEN_PLUS_SAMPLE_OUTPUT


def test_conversion(leiden_plus_sample, leiden_plus_sample_output):
    assert leiden_plus(leiden_plus_sample, indent=True) == leiden_plus_sample_output



from leidenmark import leiden_plus


def test_above():
    assert leiden_plus('\\above/') == '\\above/'
    assert leiden_plus('<= \\above/ =>') == '<ab><add place="above">above</add></ab>'


def test_interlinear():
    assert leiden_plus('||interlin: some text||') == '<add place="interlinear">some text</add>'



from leidenmark import leiden_plus


def test_paragraphos():
    assert leiden_plus('---') == '<hr/>'
    assert leiden_plus('<= --- =>') == '<ab><milestone rend="paragraphos" unit="undefined"/></ab>'
    assert (
        leiden_plus('<= 1. In between\n---\n2. Two lines =>') ==
        '<ab><l n="1">In between</l><milestone rend="paragraphos" unit="undefined"/><l n="2">Two lines</l></ab>'
    )


def test_diple():
    assert leiden_plus('((diple))') == '<milestone rend="diple" unit="undefined"/>'



class ContextInlineMixin:
    """
    A mixin for inline processors that obtains information about the element
    tree
    """

    def get_ancestors(self):
        try:
            inline_processor = self.md.treeprocessors['inline']
        except KeyError:
            raise KeyError(
                'InlineProcessor not found (ContextInlineMixin depends on it)'
            )
        else:
            return inline_processor.ancestors

    def in_leiden_plus(self):
        """
        Derives from the ancestors whether we are in a Leiden+ block
        (essentially an educated guess rn, perhaps we can make this more
        robust)
        """
        ancestors = self.get_ancestors()
        if not ancestors:
            return False
        else:
            # Perhaps we can make this more sophisticated in the future
            return ancestors[-1] == 'l' or 'ab' in ancestors



import re
from xml.etree import ElementTree as etree

from markdown.preprocessors import Preprocessor
from markdown.blockprocessors import BlockProcessor

LINE_NUM = r'([0-9]+[a-z]?´?(?:[–\-/][0-9]*[a-z]?)?)\.(-?)'


class LineNumsPreproc(Preprocessor):

    RE_REMAINDER = r'\s*(.*?)(?<!\s)\s*$'

    RE_LINE = re.compile(fr'^{LINE_NUM}{RE_REMAINDER}')
    RE_LINE_INDENT = re.compile(fr'^\((\d+),\s*indent\){RE_REMAINDER}')

    def run(self, lines):
        new_lines = []
        prev = True  # Track whether previous line has a line number or not
        # so we can separate numbered lines in blocks
        for line in lines:
            m = self.RE_LINE.search(line)
            if m:
                n, hyph, rem = m.groups()
                n = n.replace('-', '–')
                # Use en-dash for ranges instead of hyphen
                line_break = ' break=no' if hyph else ''
                if not prev:
                    new_lines.append('')
                new_lines.append(f'{{LINE n={n}{line_break}}}{rem}')
                prev = True
            else:
                m = self.RE_LINE_INDENT.search(line)
                if m:
                    n, rem = m.groups()
                    if not prev:
                        new_lines.append('')
                    new_lines.append(f'{{LINE n={n} rend=indent}}{rem}')
                    new_lines.append(f'')
                    prev = True
                else:
                    if prev:
                        new_lines.append('')
                    new_lines.append(line)
                    prev = False
        return new_lines


class NumberedBlocksProcessor(BlockProcessor):

    RE_LINE_DECL = re.compile(r'^\{LINE (.+?)\}')

    def test(self, parent, block):
        return all(
            self.RE_LINE_DECL.search(line) for line in block.split('\n')
        )

    def run(self, parent, blocks):
        lines = blocks.pop(0).split('\n')
        for line in lines:
            m = self.RE_LINE_DECL.search(line)
            attribs = [attr.split('=') for attr in m.group(1).split(' ')]
            content = line[m.end():]

            line_el = etree.SubElement(parent, 'l')
            for attr in attribs:
                line_el.set(*attr)
            line_el.text = content


def register_line_nums(md):
    md.preprocessors.register(
        LineNumsPreproc(md), 'linenums_preproc', 11
    )
    md.parser.blockprocessors.register(
        NumberedBlocksProcessor(md.parser), 'lineblocks', 120
    )



import re
import xml.etree.ElementTree as etree

from markdown.inlinepatterns import InlineProcessor

from .util import ContextInlineMixin


def format_bracket_re(opening_bracket, closing_bracket):
    re_ob = re.escape(opening_bracket)
    re_cb = re.escape(closing_bracket)
    return fr'[{re_ob}]([^{re_ob}{re_cb}\n]*?)[{re_cb}]'


RE_SUPPLIED = format_bracket_re('[', ']')
RE_ERASURE = format_bracket_re('\u301a\u27e6', '\u301b\u27e7')
RE_ABOVE = format_bracket_re('\\', '/')
RE_INTERLINEAR = r"\|\|interlin:\s*([^\|\n]+?)\|\|"


class BracketProcessor(InlineProcessor, ContextInlineMixin):

    tag = ''
    attrs = {}

    def __init__(self, *args, **kwargs):
        InlineProcessor.__init__(self, *args, **kwargs)

    def handleMatch(self, m, data):
        if not self.in_leiden_plus():
            return None, None, None
        else:
            el = etree.Element(self.tag, **self.attrs)
            el.text = m.group(1)
            return el, m.start(), m.end()


class SuppliedProcessor(BracketProcessor):

    tag = 'supplied'
    attrs = {'reason': 'lost'}


class ErasureProcessor(BracketProcessor):

    tag = 'del'
    attrs = {'rend': 'erasure'}


class AboveProcessor(BracketProcessor):

    tag = 'add'
    attrs = {'place': 'above'}


class InterlinearProcessor(InlineProcessor):

    def handleMatch(self, m, data):
        el = etree.Element('add', place='interlinear')
        el.text = m.group(1)
        return el, m.start(), m.end()


def register_brackets(md):
    md.inlinePatterns.register(
        SuppliedProcessor(RE_SUPPLIED, md), 'supplied', 5
    )
    md.inlinePatterns.register(
        ErasureProcessor(RE_ERASURE, md), 'erasure',  115
    )
    md.inlinePatterns.register(
        AboveProcessor(RE_ABOVE, md), 'above', 114
    )
    md.inlinePatterns.register(
        InterlinearProcessor(RE_INTERLINEAR, md), 'interlinear', 113
    )



from xml.etree import ElementTree as etree

from markdown.inlinepatterns import InlineProcessor

RE_HETA = r'(?:[_*]\u0323?)(h\u0323?)(?:[_*]\u0323?)'
RE_HANDSHIFT = r'\$m(\d+)'


class HetaProcessor(InlineProcessor):

    def handleMatch(self, m, data):
        el = etree.Element('em')
        el.text = m.group(1)
        return el, m.start(), m.end()


class HandShiftProcessor(InlineProcessor):

    def handleMatch(self, m, data):
        el = etree.Element('handShift', new=f'm{m.group(1)}')
        return el, m.start(), m.end()


def register_misc(md):

    md.inlinePatterns.register(
        HetaProcessor(RE_HETA, md), 'heta', 49
    )  # After UnderscoreProcessor
    md.inlinePatterns.register(
        HandShiftProcessor(RE_HANDSHIFT, md), 'handshift', 100
    )



from markdown.extensions import Extension
from markdown import markdown

from dh_utils.tei.markdown import TEIPostprocessor

from .divisions import register_divisions, deregister_paragraphs
from .line_nums import register_line_nums
from .milestones import register_milestones
from .gaps_spaces import register_gaps_spaces
from .brackets import register_brackets
from .foreign import register_foreign
from .misc import register_misc

__version__ = '0.2.1'


class LeidenPlus(Extension):

    def __init__(self, **kwargs):
        self.config = {
            'strict': [
                False,
                'Toggle strict mode (omit all Markdown specific conventions)'
            ],
            'enable_paragraphs': [False, ''],
            'indent': [False, ''],
            'with_root': [False, '']
        }
        Extension.__init__(self, **kwargs)

    def extendMarkdown(self, md):
        configs = self.getConfigs()

        md.parser.blockprocessors.deregister('olist')
        if configs['strict']:
            md.inlinePatterns.deregister('em_strong')
            md.inlinePatterns.deregister('em_strong2')

        if not configs['enable_paragraphs']:
            deregister_paragraphs(md)

        register_divisions(md)
        register_milestones(md)
        register_line_nums(md)
        register_brackets(md)
        register_gaps_spaces(md)
        register_foreign(md)
        register_misc(md)

        md.postprocessors.register(
            TEIPostprocessor(
                md, configs['indent'], configs['with_root']
            ), 'to_epidoc', 0
        )


class LeidenEscape(Extension):

    def extendMarkdown(self, md):
        md.postprocessors.register(TEIPostprocessor(md), 'to_xml', 0)


def leiden_plus(content, **kwargs):
    extensions = [LeidenPlus(**kwargs)]
    if not kwargs.get('strict', False):
        extensions.append('tables')
    return markdown(content, extensions=extensions)


def leiden_escape(content, **kwargs):
    return markdown(content, extensions=[LeidenEscape(**kwargs), 'tables'])



from leidenmark import leiden_plus


def test_line_nums():
    assert leiden_plus(
        '<=\n1. Line 1\n2. Line 2\n(3, indent) Indented line 3\n=>'
    ) == (
        '<ab><l n="1">Line 1</l>'
        '<l n="2">Line 2</l>'
        '<l n="3" rend="indent">Indented line 3</l></ab>'
    )



import re
from xml.etree import ElementTree as etree

from markdown.preprocessors import Preprocessor
from markdown.inlinepatterns import InlineProcessor

from .line_nums import LINE_NUM
from .util import ContextInlineMixin

CA_DOT = r'((?:ca\.)?)'
CA = r'((?:ca)?)'
NUM = r'(\d+(?:-\d+)?|\?|)'
LINE = r'((?:lin)?)'

RE_CHARACTER_LOST = fr'\[{CA}\.{NUM}\]'
RE_CHARACTER_ILLEGIBLE = fr'{CA}\.{NUM}'
RE_LINE_LOST = fr'lost\.{CA_DOT}{NUM}lin'
RE_SPACE = fr'vac\.{CA_DOT}{NUM}([a-zA-Z]*)'


def _handle_ca(ca, el):
    if ca:
        el.set('precision', 'low')


def _handle_num(num, el):
    if num == '?':
        el.set('extent', 'unknown')
    elif '-' in num:
        num_range = [int(i) for i in num.split('-')]
        el.set('atLeast', str(min(num_range)))
        el.set('atMost', str(max(num_range)))
    else:
        el.set('quantity', num if num else '1')


class CompleteSquareBrackets(Preprocessor):

    RE_START = re.compile(
        fr'^(?P<start>(?:{LINE_NUM})?\s*)(?P<end>[^\[\]\n]*?\])'
    )
    RE_END = re.compile(r'(?P<start>\[[^\[\]\n]*?)(?<!\s)\s*$')

    def run(self, lines):
        new_lines = []

        for line in lines:
            line = self.RE_START.sub(
                lambda m: f'{m.group("start")}[.?]' if m.group('end') == ']'
                else f'{m.group("start")}[{m.group("end")}',
                line
            )
            line = self.RE_END.sub(
                lambda m: '[.?]' if m.group() == '['
                else f'{m.group("start")}]',
                line
            )
            new_lines.append(line)

        return new_lines


class CharacterGapProcessor(InlineProcessor):

    reason = ''

    def handleMatch(self, m, data):
        ca, num = m.groups()
        el = etree.Element('gap')
        _handle_num(num, el)
        el.set('unit', 'character')
        _handle_ca(ca, el)
        if self.reason:
            el.set('reason', self.reason)
        return el, m.start(), m.end()


class CharacterLostProcessor(CharacterGapProcessor):

    reason = 'lost'


class CharacterIllegibleProcessor(CharacterGapProcessor, ContextInlineMixin):

    reason = 'illegible'

    def handleMatch(self, m, data):
        if not m.group(2) or not self.in_leiden_plus():
            return None, None, None
        else:
            return super().handleMatch(m, data)


class LineGapProcessor(InlineProcessor):

    def handleMatch(self, m, data):
        ca, num = m.groups()
        el = etree.Element('gap')
        _handle_num(num, el)
        el.set('unit', 'line')
        _handle_ca(ca, el)
        return el, m.start(), m.end()


class SpaceProcessor(InlineProcessor):

    def handleMatch(self, m, data):
        ca, num, unit = m.groups()
        el = etree.Element('space')
        _handle_num(num, el)
        if unit == 'lin':
            unit = 'line'
        elif not unit:
            unit = 'character'
        el.set('unit', unit)
        _handle_ca(ca, el)
        return el, m.start(), m.end()


def register_gaps_spaces(md):
    md.preprocessors.register(
        CompleteSquareBrackets(md), 'complete_square_brackets', 119
    )
    md.inlinePatterns.register(
        SpaceProcessor(RE_SPACE, md), 'spaces', 119
    )
    md.inlinePatterns.register(
        LineGapProcessor(RE_LINE_LOST, md), 'line_gaps', 11
    )
    md.inlinePatterns.register(
        CharacterLostProcessor(
            RE_CHARACTER_LOST, md
        ), 'character_lost', 117
    )
    md.inlinePatterns.register(
        CharacterIllegibleProcessor(
            RE_CHARACTER_ILLEGIBLE, md
        ), 'character_illegible', 4
    )



from xml.etree import ElementTree as etree

from markdown.inlinepatterns import InlineProcessor

RE_FOREIGN = r'~\|(.+?)\|~([a-zA-Z\-]+)'


class ForeignProcessor(InlineProcessor):

    def handleMatch(self, m, data):
        el = etree.Element('foreign')
        el.set('xml:lang', m.group(2))
        el.text = m.group(1)
        return el, m.start(), m.end()


def register_foreign(md):
    md.inlinePatterns.register(
        ForeignProcessor(RE_FOREIGN, md), 'foreign', 116
    )



import re
from xml.etree import ElementTree as etree

from .divisions import LEIDEN_PLUS_STATE

from markdown.blockprocessors import BlockProcessor
from markdown.inlinepatterns import InlineProcessor

RE_MIL = r'<=\.(ms|lb|cb|pb|gb) ([^<\n]+?)\s*=>'
RE_DIPLE = r'\(\(diple\??\)\)'
RE_PARAGRAPHOS = r'(?<=\n)----[^\S\n]*(?=\n)'


class MilestoneProcessor(InlineProcessor):

    def handleMatch(self, m, data):
        mil_type, content = m.groups()
        cont_parts = re.split(r'\s+', content)
        el = etree.Element('milestone')
        el.set('unit', 'section')

        if mil_type == 'ms':
            if len(cont_parts) == 1:
                if content.startswith('par.'):
                    el.set('type', 'paragraph')
                    el.set('n', content[4:])
                else:
                    el.set('n', content)
            elif len(cont_parts) == 2:
                if re.search('^[0-9.]+$', cont_parts[0]):
                    el.set('ed', cont_parts[1])
                    el.set('n', cont_parts[0])
                else:
                    el.set('ed', cont_parts[0])
                    el.set('n', cont_parts[1])
            else:
                el.set('ed', ' '.join(cont_parts[:-1]))
                el.set('n', cont_parts[-1])

        elif mil_type == 'pb':
            el.set('type', 'page')
            if len(cont_parts) == 1 and re.search(r'^.+\.page\..+$', content):
                cont_parts = content.split('.')
                el.set('ed', cont_parts[0])
                el.set('n', cont_parts[2])
            elif len(cont_parts) == 2:
                if re.search('^[0-9.]+$', cont_parts[0]):
                    el.set('ed', cont_parts[1])
                    el.set('n', cont_parts[0])
                else:
                    el.set('ed', cont_parts[0])
                    el.set('n', cont_parts[1])
            else:
                el.set('ed', cont_parts[:-1])
                el.set('n', cont_parts[-1])

        elif mil_type in ['lb', 'cb', 'gb']:
            el.set(
                'type',
                {'lb': 'line', 'cb': 'column', 'gb': 'gathering'}[mil_type]
            )
            if len(cont_parts) > 1:
                el.set('ed', cont_parts[:-1])
            el.set('n', cont_parts[-1])

        return el, m.start(), m.end()


class DipleProcessor(InlineProcessor):

    def handleMatch(self, m, data):
        el = etree.Element('milestone', rend='diple', unit='undefined')
        return el, m.start(), m.end()


class ParagraphosProcessor(BlockProcessor):

    def test(self, parent, block):
        if not re.search('^-{3,}$', block.strip()):
            return False

        try:
            state = self.parser.state[-1]
        except IndexError:
            return False
        else:
            return state == LEIDEN_PLUS_STATE

    def run(self, parent, blocks):
        blocks.pop(0)
        etree.SubElement(
            parent, 'milestone', rend='paragraphos', unit='undefined'
        )


def register_milestones(md):
    # NB: MilestoneProcessor is not official Leiden+!
    md.inlinePatterns.register(
        MilestoneProcessor(RE_MIL, md), 'milestones', 120
    )
    md.inlinePatterns.register(
        DipleProcessor(RE_DIPLE, md), 'diple', 120
    )
    md.parser.blockprocessors.register(  # Before HRProcessor
        ParagraphosProcessor(md.parser), 'paragraphos', 51
    )



from setuptools import setup, find_packages

with open('README.md') as f:
    README = f.read()

with open('requirements.txt') as f:
    REQUIREMENTS = f.readlines()

setup(
    name='leidenmark',
    description=(
        'A markdown extension for converting Leiden+ epigraphic '
        'text to TEI Epidoc XML'
    ),
    packages=find_packages(),
    install_requires=REQUIREMENTS,
    keywords='Leiden+ Markdown TEI Epidoc XML HTML',
    long_description=README,
    long_description_content_type='text/markdown',
    url='https://github.com/andredelft/leidenmark',
    python_requires='>=3.6',
    author='André van Delft',
    author_email='andre@delve.nu',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Programming Language :: Python :: 3'
    ],
    entry_points={
        'markdown.extensions': [
            'leiden_plus = leidenmark:LeidenPlus',
            'leiden_escape = leidenmark:LeidenEscape'
        ]
    }
)
